unit DHTML;

interface

uses Classes, Contnrs, SysUtils, Dialogs, Forms, Controls;

var
  DebugLines: TStrings;
  
type
  TColor = $000000..$FFFFFF;

  TStrArray = array of String;

  TDHTMLFont = record
    Size: integer;
    Face: string;
    Color: TColor;
  end;

  TInputType = (itUnknown, itButton, itCheckBox, itFile,
                itHidden, itPassword, itRadio, itReset,
                itSubmit, itText);

  TDHTMLDocument = class(TStringList)
    private
//      procedure LoadFromFile(const FileName: string); reintroduce; virtual; abstract;
//      procedure SaveToFile(const FileName: string); reintroduce; virtual; abstract;
  end;

  TDHTMLTagList = class;

  TDHTMLTag = Class(TObject)
  private
    FTag: string;
    FItems: TDHTMLTagList;
    FTagText: string;
    FInnerText: string;
    FOuterText: string;
    function GetItems(i: integer): TDHTMLTag;
    procedure SetItems(i: integer; const Value: TDHTMLTag);
    function GetItemCount: Integer;
  public
    constructor Create;
    property Tag: string read FTag write FTag;
    property ItemCount: integer read GetItemCount;
    property Items[i: integer]: TDHTMLTag read GetItems write SetItems; default;
    property TagText: string read FTagText write FTagText;
    property InnerText: string read FInnerText write FInnerText;
    property OuterText: string read FOuterText write FOuterText;
    procedure Assign(AItem: TDHTMLTag); reintroduce;
    procedure ClearItems;
    procedure AddItem(Item: TDHTMLTag);
    function RemoveItem(Item: TDHTMLTag): integer; overload;
    function RemoveItem(Item: TDHTMLTag; var Parent: TDHTMLTag): integer; overload;
    function ItemIndex(Item: TDHTMLTag): integer;
    procedure DeleteItem(Index: integer);
    procedure InsertItem(Index: integer; Item: TDHTMLTag);
    procedure SetAttribute(Attribute, Value: string);
    procedure SetColor(Value: integer);
    function ReplaceItem(Target, Sorce: TDHTMLTag): boolean;
    function AddFont(Font: TDHTMLFont): TDHTMLTag;
    function GetAttribute(Attribute: string): string;
    function GetAttributeInt(Attribute: string): Integer;
    function GetColor: integer;
    function GetFont: TDHTMLFont;
    function GetElementByName(Name: string): TDHTMLTag;
    function GetElementByID(ID: integer): TDHTMLTag;
    function GetElementByTag(Tag: string; SubLayers: Integer = high(integer)): TDHTMLTag;
    function GetElementsByTag(Tag: string; SubLayers: Integer = high(integer)): TDHTMLTagList;
    function GetElementsByAttribute(Attribute, Value: string; SubLayers: Integer = high(integer)): TDHTMLTagList;
  end;

  TDHTMLTableRow = class;

  TRowCompare = function(row1, row2: TDHTMLTag): integer of object;

  TDHTMLTable = class(TDHTMLTag)
  private
    FRowCount: integer;
    FDoCompare: TRowCompare;
  protected
    function GetRow(Index: Integer): TDHTMLTag;
  published
  public
    constructor Create;
    procedure SortRows;
    //procedure Assign(AItem: TDHTMLTag); reintroduce;      
    function SwapRows(row1,row2: TDHTMLTag): boolean;
    function GetRowCount: integer;
    property DoCompare: TRowCompare read FDoCompare write FDoCompare;
    property RowCount: integer read FRowCount;
    property Rows[Index: integer]: TDHTMLTag read GetRow; default;
  end;

  TDHTMLTableRow = class(TDHTMLTag)
  protected
    function GetCell(Index: Integer): TDHTMLTag;
  published
  public
    constructor Create;
    //procedure Assign(AItem: TDHTMLTag); reintroduce;
    property Cells[Index: integer]: TDHTMLTag read GetCell; default;
  end;

  TDHTMLTagList = class(TObjectList)
  protected
    function GetItem(Index: Integer): TDHTMLTag;
    procedure SetItem(Index: Integer; AObject: TDHTMLTag);
  public
    constructor Create;
    property Items[Index: Integer]: TDHTMLTag read GetItem write SetItem; default;
  end;

  TDHTMLObject = class(TDHTMLTag)
  private           
    FItems: TObject;
  protected
    property Items: TObject read FItems;
  public
  end;

  TDHTMLInput = class(TDHTMLObject)
  private
    FInputType: TInputType;
    FValue: string;
    FTyp: string;
    FName: string;
    FID: string;
    procedure SetValue(const Value: string);
    function GetValue: string;
    function GetTyp: string;
    procedure SetTyp(const Value: string);
    function GetName: string;
    procedure SetName(const Value: string);
    function GetID: string;
    procedure SetID(const Value: string);
  public
    constructor Create(InputType: TInputType);
    property InputType: TInputType read FInputType;
    property Typ: string read GetTyp write SetTyp;
    property Name: string read GetName write SetName;
    property ID: string read GetID write SetID;
    property Value: string read GetValue write SetValue;
  end;

  TDHTML = class(TObjectList)
  private
    IntDocument: TDHTMLDocument;
    FDocument: TDHTMLDocument;
    FMainTag: TDHTMLTag;
    procedure SetDocument(const Value: TDHTMLDocument);
    procedure MakeDoc;
    procedure BuildList;
  public
    constructor Create;
    procedure LoadFromFile(const FileName: string);
    procedure SaveToFile(const FileName: string);
    property MainTag: TDHTMLTag read FMainTag write FMainTag;
    property Document: TDHTMLDocument read FDocument write SetDocument;
    function GetDocumentString: string;
  end;

implementation

procedure WriteBug(bug: string);
begin
  if assigned(DebugLines) then begin
    DebugLines.Add(bug);
  end;
end;

const
  OpenTags: array[0..4] of String = ('!', 'br', 'img', 'meta', 'link');
  UnClosedTags: array[0..1] of String = ('p', 'span');
  InlineTags: array[0..36] of String = ('/a', '/abbr', '/acronym', '/applet',
  '/b', '/basefont', '/bdo', '/big', '/br', '/button', '/cite', '/code', '/del',
  '/dfn', '/em', '/font', '/i', '/img', '/ins', '/input', '/iframe', '/kbd',
  '/label', '/map', '/object', '/q', '/samp', '/script', '/select', '/small',
  '/span', '/strong', '/sub', '/sup', '/textarea', '/tt', '/var');
  InputStrings: array[TInputType] of String = ('unknown', 'button', 'checkBox', 'file',
                'hidden', 'password', 'radio', 'reset',
                'submit', 'text');

function StrIsIn(value: string; arr: array of String): boolean;
var
  i: integer;
begin
  result:= false;
  for i:= 0 to high(arr) do begin
    if value = arr[i] then begin
      result:= true;
      break;
    end;
  end;
end;

function CaseOfString(value: string; arr: array of String): integer;
var
  i: integer;
begin
  result:= 0;
  for i:= 0 to high(arr) do begin
    if value = arr[i] then begin
      result:= i;
      break;
    end;
  end;
end;
{ TDHTML }

procedure TDHTML.BuildList;
var
  index, p: integer;
  tmps: string;
  LastUnclosed: string;
  curs: Integer;

  function MakeTagText(text: string): string;
  var
    tmps: string;
    p: integer;
  begin
    //text:= LowerCase(text);
    p:= pos('>', text);
    if p<>0 then begin
      tmps:= copy(text, 1, p);
      result:= '<' + tmps;
    end;
  end;

  function MakeText(tag: string): string;
  var
    p: integer;

    function CutSpaces(s: string): string;
    var
      i: integer;
      l: integer;
      c: char;
    begin
      result:= '';
      for i := 1 to length(s) do begin
        c:= s[i];
        l:= length(result);
        if l > 0 then begin
          if (result[l] <> c) or(result[l] <> ' ') then begin
            result:= result + c;
          end;
        end else result:= c;
      end;
    end;

  begin
    tag:= CutSpaces(tag);
    p:= pos('>', tag);
    if p<>0 then begin
      Result:= copy(tag, p+1, high(integer));
    end else result:= '';
  end;

  function WorkThemTags(tag: string): TDHTMLTag;
  var
    tmps, tmps2: string;
    p: integer;
  begin
    tmps:= '';
    tag:= LowerCase(tag);
    if StrIsIn(tag, UnClosedTags) then
      LastUnclosed:= tag;
    Result:= TDHTMLTag.Create;
    Result.Tag:= tag;
    Result.TagText:= MakeTagText(IntDocument[index]);
    Result.InnerText:= MakeText(IntDocument[index]);
    if not StrIsIn(tag, OpenTags) then begin
      inc(index);
      while (index < IntDocument.Count-1) do begin
        p:= pos('>', IntDocument[index]);
        if p<>0 then begin
          tmps:= copy(IntDocument[index], 1, p-1);
          if tmps[1] = '!' then
            tmps:= tmps[1];
          p:= pos(' ', tmps);
          if p <> 0 then begin
            tmps:= copy(tmps, 1, p-1);
           end;
         end else begin
          inc(index);
          continue;
        end;
        tmps:= LowerCase(tmps);
        tmps2:= tmps;
        if tmps2[1] <> '/' then
          tmps2:= '/'+tmps2;
        if ((LastUnclosed <> '') and (not StrIsIn(tmps2, InlineTags))) then begin
          Result.OuterText:= MakeText(IntDocument[index]);
          LastUnclosed:= '';
          dec(index);
          exit;
        end;
        if (tmps = '/'+tag) or (tmps = '/'+tag+'>') then begin
          Result.OuterText:= MakeText(IntDocument[index]);
          exit;
         end else begin
          if tmps[1]<>'/' then begin
            Result.AddItem(WorkThemTags(tmps));
           end else begin
            //is doof-.-
            sleep(1);
          end;
        end;
        inc(index);
      end;
    end;
  end;

begin
  curs:= screen.cursor;
  screen.Cursor:= crHourGlass;
  Application.ProcessMessages;
  LastUnclosed:= '';
  index:= 0;
  while index < IntDocument.Count-2 do begin
    if IntDocument[index][1] <> '/' then begin
      p:= pos('>', IntDocument[index]);
      if p<>0 then begin
        tmps:= copy(IntDocument[index], 1, p-1);
        if tmps[1] = '!' then
          tmps:= tmps[1];
        p:= pos(' ', tmps);
        if p <> 0 then begin
          tmps:= copy(tmps, 1, p-1);
        end;
       end else begin
        inc(index);
        continue;
      end;
      MainTag.AddItem(WorkThemTags(tmps));
    end;
    inc(index);
  end;
  Screen.Cursor:= curs;
end;

constructor TDHTML.Create;
begin
  inherited;
  MainTag:= TDHTMLTag.Create;
  FDocument:= TDHTMLDocument.Create;
  IntDocument:= TDHTMLDocument.Create;
  IntDocument.Delimiter:= '<';
  IntDocument.StrictDelimiter:= True;
end;

function TDHTML.GetDocumentString: string;
var
  mesh: string;
  i: integer;
  tmpl: TStringList;

  procedure writetag(tag: TDHTMLTag);
  var
    j: integer;
    tmpmesh: string;
  begin
    tmpl.Add(mesh + tag.TagText);
    tmpmesh:= mesh;
    mesh:= mesh+' ';
    if (tag.InnerText <> '') and (tag.InnerText <> ' ')then begin
      tmpl.Add(tag.InnerText);
    end;
    for j := 0 to tag.ItemCount - 1 do begin
      writetag(tag.Items[j]);
    end;
    mesh:= tmpmesh;
    if not StrIsIn(tag.Tag, OpenTags) then begin
      tmpl.Add(mesh + '</'+tag.tag+'>');
    end;
    if (tag.OuterText <> '') and (tag.OuterText <> ' ')then begin
      tmpl.Add(tag.OuterText);
    end;
  end;

begin
  tmpl:= TStringList.Create;
  mesh:= '';
  for i := 0 to MainTag.ItemCount - 1 do begin
    writetag(MainTag.Items[i]);
  end;
  Result:= tmpl.Text;
end;

procedure TDHTML.LoadFromFile(const FileName: string);
begin
  FMainTag.ClearItems;
  FDocument.Clear;
  (FDocument as TStringList).LoadFromFile(FileName);
  MakeDoc;
  BuildList;
end;


procedure TDHTML.MakeDoc;
var
  s: string;
  i: integer;
  curs: integer;
begin
  curs:= screen.cursor;
  screen.Cursor:= crHourGlass;
  Application.ProcessMessages;
  IntDocument.Clear;
  s:= FDocument.Text;
  if s <> '' then
    sleep(1);
  for I := 0 to FDocument.Count-1 do begin
    IntDocument.Add(TrimLeft(FDocument[i]));
  end;
  s:= IntDocument.Text;
  s:= StringReplace(s, #$0D, '', [rfReplaceAll]);
  s:= StringReplace(s, '>'#$0A, '>', [rfReplaceAll]);
  s:= StringReplace(s, #$0A'<', '<', [rfReplaceAll]);
  s:= StringReplace(s, #$0A, ' ', [rfReplaceAll]);
  s:= StringReplace(s, #$09, '', [rfReplaceAll]);
  IntDocument.DelimitedText:= s;
  for I := IntDocument.Count - 1 downto 0 do begin
    if IntDocument[i] = '' then begin
      IntDocument.Delete(i);
    end;
  end;
  Screen.Cursor:= curs;
end;

procedure TDHTML.SaveToFile(const FileName: string);
var
  tmpl: TStringList;
begin
  tmpl:= TStringList.create; 
  tmpl.Text:= GetDocumentString;
  tmpl.SaveToFile(FileName);
end;

procedure TDHTML.SetDocument(const Value: TDHTMLDocument);
begin
  FDocument.Clear;
  FDocument:= Value;
  MakeDoc;
  BuildList;
end;

{ TDHTMLInput }

constructor TDHTMLInput.Create(InputType: TInputType);
begin
  inherited Create;
  FInputType:= InputType;
  FTyp:= '';
  FValue:= '';
  FName:= '';
  FID:= '';
// more to come
// mebby
end;

function TDHTMLInput.GetID: string;
begin
  FID:= GetAttribute('id');
  Result:= FID;
end;

function TDHTMLInput.GetName: string;
begin
  FName:= GetAttribute('name');
  Result:= FName;
end;

function TDHTMLInput.GetTyp: string;
begin
  FTyp:= GetAttribute('type');
  FInputType:= TInputType(CaseOfString(FTyp, InputStrings));
  Result:= FTyp;
end;

function TDHTMLInput.GetValue: string;
begin
  FValue:= GetAttribute('value');
  Result:= FValue;
end;

procedure TDHTMLInput.SetID(const Value: string);
begin
  FID:= Value;
  SetAttribute('id', Value);
end;

procedure TDHTMLInput.SetName(const Value: string);
begin
  FName := Value;
  SetAttribute('name', Value);
end;

procedure TDHTMLInput.SetTyp(const Value: string);
begin
  FTyp := Value;
  FInputType:= TInputType(CaseOfString(FTyp, InputStrings));
  SetAttribute('type', Value);
end;

procedure TDHTMLInput.SetValue(const Value: string);
begin
  FValue := Value;
  SetAttribute('value', Value);
end;

{ TDHTMLTag }

function TDHTMLTag.GetAttribute(Attribute: string): string;
var
  p: integer;
begin
  Result:= lowercase(FTagText);
  Attribute:= lowercase(Attribute);
  p:= AnsiPos(Attribute+'=', Result);
  if p <> 0  then begin
    Result:= Copy(FTagText, p, High(integer));
    p:= AnsiPos('"', Result);
    if p <> 0  then begin
      Result:= Copy(Result, p+1, High(integer));
      p:= AnsiPos('"', Result);
      if p <> 0  then begin
        Result:= Copy(Result, 1, p-1);
       end else begin
        Result:= '';
      end;
     end else begin
      Result:= '';
    end;
   end else begin
    Result:= '';
  end;
end;

function TDHTMLTag.GetAttributeInt(Attribute: string): Integer;
var
  p: integer;
  res: string;
begin
  Res:= lowercase(FTagText);
  Attribute:= lowercase(Attribute);
  p:= AnsiPos(Attribute+'=', Res);
  if p <> 0  then begin
    Res:= Copy(FTagText, p, High(integer));
    p:= AnsiPos('"', Res);
    if p <> 0  then begin
      Res:= Copy(Res, p+1, High(integer));
      p:= AnsiPos('"', Res);
      if p <> 0  then begin
        Res:= Copy(Res, 1, p-1);
       end else begin
        Res:= '';
      end;
     end else begin
      Res:= '';
    end;
   end else begin
    Res:= '';
  end;
  Result:= 0;
  TryStrToInt(res, Result);
end;

function TDHTMLTag.GetColor: integer;
var
  tmp: string;
begin
  Result:= 0;
  tmp:= GetAttribute('color');
  tmp:= StringReplace(tmp, '#', '$', [rfReplaceAll]);
  TryStrToInt(tmp, Result);
end;

function TDHTMLTag.GetElementByID(ID: integer): TDHTMLTag;
var
  tmp: TDHTMLTagList;
begin
  tmp:= GetElementsByAttribute('id', inttostr(id));
  Result:= tmp[0];
  tmp.free;
end;

function TDHTMLTag.GetElementByName(Name: string): TDHTMLTag;
var
  tmp: TDHTMLTagList;
begin
  tmp:= GetElementsByAttribute('name', name);
  Result:= tmp[0];
  tmp.free;
end;

function TDHTMLTag.GetElementByTag(Tag: string;
  SubLayers: Integer): TDHTMLTag;
var
  tmpl: TDHTMLTagList;
begin
  tmpl:= GetElementsByTag(tag,SubLayers);
  if tmpl.Count > 0 then begin
    result:= tmpl[0];
   end else begin
    result:= nil;
  end;
  tmpl.Free;
end;

function TDHTMLTag.GetElementsByAttribute(Attribute, Value: string; SubLayers: Integer = high(integer)): TDHTMLTagList;
var
  i,j: integer;
  tmp: TDHTMLTagList;
  tmptag: TDHTMLtag;
  tmps: string;
begin
  Result:= TDHTMLTagList.Create;
  try
  for I := 0 to FItems.Count - 1 do begin
    if sublayers = 2147483634 then begin
      sleep(1);
    end;
  //  WriteDebug(' ');
   // WriteDebug('sub=' + inttostr(SubLayers));
    tmptag:= Fitems[i];
    tmps:= tmptag.GetAttribute(Attribute);
   // WriteDebug('tag= ' + tmptag.tag);
    //WriteDebug('attr= ' + tmps);
    if tmps = value then begin
      Result.Add(FItems[i]);
    end;
    if SubLayers > 0 then begin
      tmp:= FItems[i].GetElementsByAttribute(Attribute, Value, SubLayers-1);
     // WriteDebug(' ');
      //WriteDebug('sub=' + inttostr(SubLayers));
      //WriteDebug('tmp cnt= ' + IntToStr(tmp.Count));
      for j := 0 to tmp.Count - 1 do begin
        Result.Add(tmp[j]);
      end;
      if assigned(tmp) then
        tmp.free;
    end;
  end;
   except
  //   WriteDebug('!!!!!!!!!!ERROR!!!!!!!!');
  end;
end;             

function TDHTMLTag.GetElementsByTag(Tag: string; SubLayers: Integer = high(integer)): TDHTMLTagList;
var
  i, j: integer;
  tmp: TDHTMLTagList;
begin
  Result:= TDHTMLTagList.Create;
  for I := 0 to FItems.Count - 1 do begin
    if Fitems[i].Tag = Tag then begin
      Result.Add(FItems[i]);
    end;
    if SubLayers > 0 then begin
      tmp:= FItems[i].GetElementsByTag(Tag, SubLayers-1);
      for j := 0 to tmp.Count - 1 do begin
        Result.Add(tmp[j]);
      end;
    tmp.free;
    end;
  end;
end;

function TDHTMLTag.GetFont: TDHTMLFont;
var
  tmp: TDHTMLTagList;
begin
  tmp:= GetElementsByTag('font');
  if assigned(tmp[0]) then begin
    Result.Face:= tmp[0].GetAttribute('face');
    Result.Size:= tmp[0].GetAttributeInt('size');
    Result.Color:= tmp[0].GetColor;
  end;
  tmp.free;
end;

procedure TDHTMLTag.SetAttribute(Attribute, Value: string);
var
  tmps, tmps2: string;
  p: integer;
begin
  Attribute:= lowercase(Attribute);
  tmps2:= '';
  tmps:= lowercase(FTagText);
  p:= pos(Attribute, tmps);
  if p <> 0 then begin
    tmps2:= copy(FTagText, p, high(integer));
    tmps:= copy(FTagText, 1, p-1);
    p:= pos('"', tmps2);
    tmps2:= copy(tmps2, p+1, high(integer));
    p:= pos('"', tmps2);
    tmps2:= copy(tmps2, p+1, length(tmps2)-(p+1));
  end else begin
    tmps:= copy(FTagText, 1, length(tmps)-1);
  end;
  tmps:= tmps + ' ' + Attribute + '="' + Value + '"';
  tmps:= tmps + tmps2 + '>';
  FTagText:= tmps;
end;

procedure TDHTMLTag.SetColor(Value: integer);
var
  tmps, tmps2: string;
  p: integer;
  Attribute: string;
begin
  Attribute:= 'color';
  tmps:= lowercase(FTagText);
  p:= pos(Attribute, tmps);
  if p <> 0 then begin
    tmps2:= copy(tmps, p-1, high(integer));
    p:= pos('"', tmps2);
    tmps2:= copy(tmps2, p+1, high(integer));
  end;
  tmps:= copy(tmps, 1, Length(tmps)-1);
  tmps:= tmps + ' ' + Attribute + '="#' + IntToHex(value,6) + '"';
  tmps:= tmps + tmps2 + '>';
  FTagText:= tmps;
end;

function TDHTMLTag.AddFont(Font: TDHTMLFont): TDHTMLTag;
begin
  Result:= TDHTMLTag.Create;
  Result.Tag:= 'font';
  Result.TagText:= '<font>';
  Result.SetAttribute('size', IntToStr(Font.Size));
  Result.SetColor(Font.color);
  Result.SetAttribute('face', Font.face);
  AddItem(Result);
end;

procedure TDHTMLTag.AddItem(Item: TDHTMLTag);
begin
  FItems.Add(Item)
end;

procedure TDHTMLTag.Assign(AItem: TDHTMLTag);
begin
  if assigned(AItem) then begin
    FTag:= AItem.Tag;
    FTagText:= AItem.TagText;
    if Assigned(Fitems) then FItems.Free;

    FItems:= AItem.FItems;
    FInnerText:= AItem.InnerText;
    FOuterText:= AItem.OuterText;
   end else begin
    FItems:= NIL;
    FTag:= '';
    FTagText:= '';
    FInnerText:='';
    FOuterText:= '';
  end;
end;

procedure TDHTMLTag.ClearItems;
begin
  FItems.Clear;
end;

constructor TDHTMLTag.Create;
begin
  inherited;
  FItems:= TDHTMLTagList.Create;
  FItems.OwnsObjects:= true;
  FTag:= '';
  FTagText:= '';
  FInnerText:= '';
  FOuterText:= '';
end;

procedure TDHTMLTag.DeleteItem(Index: integer);
begin
  FItems.Delete(Index);
end;

procedure TDHTMLTag.InsertItem(Index: integer; Item: TDhtmlTag);
begin
  FItems.Insert(Index, Item);
end;

function TDHTMLTag.ItemIndex(Item: TDHTMLTag): integer;
begin
  if Assigned(item) then begin
    result:= FItems.IndexOf(item);
   end else begin
    Result:= -1;
  end;
end;

function TDHTMLTag.RemoveItem(Item: TDHTMLTag): integer;
var
  rem: integer;
  i: integer;
begin
  rem:= FItems.Remove(item);
  if rem < 0 then begin
    for I := 0 to FItems.Count-1 do begin
      rem:= (Fitems[i] as TDHTMLTag).RemoveItem(Item);
      if rem >= 0 then
        break;
    end;
  end;
  result:= rem;
end;

function TDHTMLTag.RemoveItem(Item: TDHTMLTag; var Parent: TDHTMLTag): integer;
var
  rem: integer;
  i: integer;
begin
  rem:= FItems.Remove(item);
  if rem < 0 then begin
    for I := 0 to FItems.Count-1 do begin
      rem:= (Fitems[i] as TDHTMLTag).RemoveItem(Item, parent);
      if rem >= 0 then
        break;
    end;
   end else begin
    parent:= self;
  end;
  result:= rem;
end;

function TDHTMLTag.ReplaceItem(Target, Sorce: TDHTMLTag): boolean;
var
 index: integer;
 parent: TDHTMLTag;
begin
  Result:= false;
  index:= RemoveItem(Target, parent);
  if index >= 0 then begin
    parent.InsertItem(Index, Sorce);
    Result:= true;
  end;
end;

function TDHTMLTag.GetItemCount: Integer;
begin
  if Assigned(FItems) then begin
    Result:= FItems.count;
   end else begin
    Result:= -1;
  end;
end;

procedure TDHTMLTag.SetItems(i: integer; const Value: TDHTMLTag);
begin
  FItems[i]:= Value;
end;

function TDHTMLTag.GetItems(i: integer): TDHTMLTag;
begin
  Result:= (FItems[i] as TDHTMLTag);
end;

{ TDHTMLTagList }

constructor TDHTMLTagList.Create;
begin
  inherited;
  OwnsObjects:= false;
end;

function TDHTMLTagList.GetItem(Index: Integer): TDHTMLTag;
begin
  if Index < Count then begin
    Result:= (inherited Items[Index]) as TDHTMLTag;
   end else begin
    result:= nil;
  end;
end;

procedure TDHTMLTagList.SetItem(Index: Integer; AObject: TDHTMLTag);
begin
  inherited Items[Index]:= AObject;
end;

{ TDHTMLTable }
{
procedure TDHTMLTable.Assign(AItem: TDHTMLTag);
begin
  FTag:= AItem.Tag;
  FTagText:= AItem.TagText;
  if Assigned(Fitems) then FItems.Free;

  FItems:= AItem.FItems;
  FInnerText:= AItem.InnerText;
  FOuterText:= AItem.OuterText;
end;    }

constructor TDHTMLTable.Create;
begin
  inherited;
  GetRowCount;
  FDoCompare:= nil;
end;


function TDHTMLTable.GetRow(Index: Integer): TDHTMLTag;
var
  tmp: TDHTMLTagList;
begin
  //Result:= TDHTMLTableRow.Create;
  tmp:= GetElementsByTag('tr', 0);   
  if tmp.Count > 0 then begin
    Result:= tmp[index];
   end else begin
    Result:= nil;
  end;
  tmp.Free;
end;

{ TDHTMLRow }
{
procedure TDHTMLTableRow.Assign(AItem: TDHTMLTag);
begin
  FTag:= AItem.Tag;
  FTagText:= AItem.TagText;
  if Assigned(Fitems) then FItems.Free;

  FItems:= AItem.FItems;
  FInnerText:= AItem.InnerText;
  FOuterText:= AItem.OuterText;
end;      }

constructor TDHTMLTableRow.Create;
begin
  inherited;
  FTag:= 'tr';
  FTagText:= '<tr>';
end;

function TDHTMLTableRow.GetCell(Index: Integer): TDHTMLTag;
var
  tmp: TDHTMLTagList;
begin
  tmp:= GetElementsByTag('td', 1);
  if tmp.Count > 0 then begin
    Result:= tmp[index];
   end else begin
    Result:= nil;
  end;
  tmp.Free;
end;

function TDHTMLTable.GetRowCount: integer;
var
  tmp: TDHTMLTagList;
begin
  tmp:= GetElementsByTag('tr', 0);
  FRowCount:= tmp.Count;
  tmp.free;
  result:= FRowCount;
end;

procedure TDHTMLTable.SortRows;
var
  i: integer;
  sorted: Boolean;
  r1, r2: TDHTMLTag;
  cnt: integer;
begin
  repeat
    sorted := true;
    i := 0;
    r1:= Rows[i];
    r2:= Rows[i+1];
   // cnt:= GetRowCount -1;
    //writebug('cnt: ' + inttostr(cnt));
    while assigned (r2) do begin
      if DoCompare(r1, r2) > 0  then begin //tauschen
        sorted:= not SwapRows(r1, r2);
      end;
      inc(i);
      r1 := r2;
      cnt:= GetRowCount -1;
      //writebug('cnt: ' + inttostr(cnt));
      if i >= cnt then begin
        r2:= nil;
       end else begin
        r2:= Rows[i+1];
      end;
    end;
  until sorted;
end;

function TDHTMLTable.SwapRows(row1, row2: TDHTMLTag): boolean;
var
  i1, i2: Integer;
  r1,r2: TDHTMLTag;
begin
  FItems.OwnsObjects:= false;
  Result:= false;
  i1:= FItems.IndexOf(row1);
  //writebug('itemindex1 holen: ' + inttostr(i1));
  //writebug(inttostr(GetRowCount-1));
  i2:= FItems.IndexOf(row2);
  //writebug('itemindex2 holen: ' + inttostr(i2));
  //writebug(inttostr(GetRowCount-1));
  if (i1 >= 0) and (i2 >= 0) then begin
    r1:= FItems[i1];
    r2:= FItems[i2];
    //writebug('item1 in r');
    //writebug(inttostr(GetRowCount-1));
    FItems[i1]:= r2;
    //writebug('item2 auf 1 schieben');
    //writebug(inttostr(GetRowCount-1));
    FItems[i2]:= r1;
    //writebug('r auf 2 schieben');
    //writebug(inttostr(GetRowCount-1));
    Result:= true;
  end;
  FItems.OwnsObjects:= true;
end;

end.
